from collections import namedtuple
from . import _pgmsg
from ._utils import get_exc_from_msg
from ._exceptions import ProgrammingError, InterfaceError
from ._transaction import Transaction
from ._cursor import Cursor


NO_EXECUTE = object()


ParameterDesc = namedtuple('ParameterDesc', ['name', 'type'])


class PreparedStatement:
    def __init__(self, conn, query):
        self.conn = conn
        self.query = query

        # even though the null character is valid UTF-8, we can't use
        # it in queries, because at the protocol level, the queries
        # are sent as zero-terminated strings.
        if '\x00' in self.query:
            raise ProgrammingError(
                'NULL character is not valid in PostgreSQL queries.')

        self._execute_started = False
        self._portal_closed = False

        self._initialized = False
        self._initial_transaction = None
        self._param_oids = None
        self._row_desc = None
        self._stmt_name = None
        self._query_row_count = None

    async def _init(self):
        if self._initialized:
            return

        if self.conn.closed.is_set():
            raise ProgrammingError('Connection is closed.')

        if not self.conn._pg_types_loaded.is_set():
            # the "if" is not technically necessary, but avoids an
            # extra trio checkpoint.
            await self.conn._pg_types_loaded.wait()

        try:
            await self._parse_query()
        finally:
            await self.conn._wait_for_ready()

        self._initialized = True

        return self

    async def _parse_query(self):
        self._stmt_name = self.conn._get_unique_id('stmt')

        parse_msg = _pgmsg.Parse(self._stmt_name, self.query)
        describe_stmt_msg = _pgmsg.Describe(b'S', self._stmt_name)
        sync_msg = _pgmsg.Sync()
        await self.conn._send_msg(
            parse_msg, describe_stmt_msg, sync_msg)

        while True:
            msg = await self.conn._get_msg(_pgmsg.ParseComplete,
                                           _pgmsg.ErrorResponse,
                                           _pgmsg.RowDescription,
                                           _pgmsg.ParameterDescription,
                                           _pgmsg.NoData)
            if isinstance(msg, _pgmsg.ErrorResponse):
                # DESCRIBE command _can_ return an error, but only if
                # the portal/prepared statement specified does not
                # exist. In our case, this should never happen.
                raise get_exc_from_msg(
                    msg,
                    desc_prefix=(
                        f'Error preparing statement: {self.query}\n   '
                    ),
                )
            elif isinstance(msg, _pgmsg.ParameterDescription):
                self._param_oids = msg.param_oids
            elif isinstance(msg, _pgmsg.RowDescription):
                # this description is incomplete (since protocol
                # format is only known after bind, but can still be
                # used to return the list of parameters and their
                # types in the "parameters" proptery)
                self._row_desc = msg.fields
                break
            elif isinstance(msg, _pgmsg.NoData):
                self._row_desc = None
                break

        msg = await self.conn._get_msg(_pgmsg.ReadyForQuery)
        await self.conn._handle_msg_ready_for_query(msg)

    @property
    def parameters(self):
        if not self._row_desc:
            return None

        return [
            ParameterDesc(
                name=name,
                type=self.conn._codec_helper._oid_to_name.get(type_oid)
            )
            for name, table_oid, column_attr_id, type_oid, type_len,
                type_modifier, format_code
            in self._row_desc
        ]

    @property
    def rowcount(self):
        return self._query_row_count

    async def cursor(self, *params, **kwargs):
        await self.execute(limit=NO_EXECUTE)
        cursor = Cursor(self, **kwargs)
        await cursor._init()
        return cursor

    async def execute(self, *params, limit=None):
        if not self._initialized:
            raise ProgrammingError(
                'PreparedStatement not initialized')

        if self.conn.closed.is_set():
            raise ProgrammingError('Connection is closed.')

        if not self.conn._pg_types_loaded.is_set():
            # the "if" is not technically necessary, but avoids an
            # extra trio checkpoint.
            await self._pg_types_loaded.wait()

        if (limit and limit is not NO_EXECUTE and
            not self.conn.in_transaction):
            # we don't allow this because the only use case for
            # "limit" is to use exec_continue later, but exec_continue
            # won't work outside a transaction (postgres will complain
            # that "portal does not exist")
            raise InterfaceError(
                'limit parameter not supported outside a transaction')

        if self._execute_started and not self._portal_closed:
            # close the previously running portal
            await self._close_portal()
            await self.conn._wait_for_ready()

        self._initial_transaction = self.conn.current_transaction
        try:
            results = await self._execute(*params, limit=limit)
        finally:
            await self.conn._wait_for_ready()

        return results

    async def _execute(self, *params, limit=None):
        self._execute_started = True
        self._portal_closed = False
        self._portal_name = self.conn._get_unique_id('portal')

        should_close = False
        protocol_format = self.conn.protocol_format

        enc_params = []
        for param, param_oid in zip(params, self._param_oids):
            param_value = self.conn._codec_helper.encode_value(
                param, param_oid,
                protocol_format=protocol_format)
            enc_params.append(param_value)

        bind_msg = _pgmsg.Bind(self._portal_name, self._stmt_name,
                               params=enc_params,
                               param_format_codes=[protocol_format],
                               result_format_codes=[protocol_format])
        describe_portal_msg = _pgmsg.Describe(b'P', self._portal_name)
        sync_msg = _pgmsg.Sync()
        if limit is not NO_EXECUTE:
            execute_msg = _pgmsg.Execute(self._portal_name,
                                         max_rows=limit or 0)
            msgs = [bind_msg, describe_portal_msg, execute_msg,
                    sync_msg]
        else:
            msgs = [bind_msg, describe_portal_msg, sync_msg]
        await self.conn._send_msg(*msgs)

        # sent messages, so no other query allowed until we have a
        # ReadyForQuery
        self.conn._is_ready = False

        results = []
        while True:
            msg = await self.conn._get_msg(
                _pgmsg.ErrorResponse,
                _pgmsg.BindComplete,
                _pgmsg.RowDescription,
                _pgmsg.NoData,
                _pgmsg.EmptyQueryResponse,
                _pgmsg.CommandComplete,
                _pgmsg.DataRow,
                _pgmsg.ErrorResponse,
                _pgmsg.PortalSuspended)
            if isinstance(msg, _pgmsg.ErrorResponse):
                raise get_exc_from_msg(
                    msg,
                    desc_prefix=(
                        f'Error executing statement: {self.query}\n   '
                    ),
                )
            elif isinstance(msg, _pgmsg.DataRow):
                row = self.conn._codec_helper.decode_row(
                    msg.columns, self._row_desc)
                results.append(row)
            elif isinstance(msg, _pgmsg.RowDescription):
                self._row_desc = msg.fields
                if limit is NO_EXECUTE:
                    break
            elif isinstance(msg, _pgmsg.NoData):
                should_close = True
                if limit is NO_EXECUTE:
                    break
            elif isinstance(msg, _pgmsg.EmptyQueryResponse):
                should_close = True
                break
            elif isinstance(msg, _pgmsg.CommandComplete):
                if msg.cmd_tag.startswith(b'SELECT'):
                    _, rows = msg.cmd_tag.split(b' ')
                    self._query_row_count = int(rows.decode('ascii'))
                else:
                    self._query_row_count = None
                should_close = True
                break
            elif isinstance(msg, _pgmsg.PortalSuspended):
                break

        if should_close:
            await self.conn._wait_for_ready()
            await self._close_portal()

        return results

    async def exec_continue(self, limit=None):
        if self.conn.closed.is_set():
            raise ProgrammingError('Connection is closed.')

        if not self._execute_started:
            raise ProgrammingError(
                'exec_continue can only be called after execute is '
                'called')

        if not self.conn.in_transaction:
            # we don't allow this because outside a transaction
            # postgres will complain that "portal does not exist"
            raise InterfaceError(
                'Cannot continue execution of a statement outside '
                'a transaction')

        if self.conn.current_transaction != self._initial_transaction:
            raise ProgrammingError(
                'Cannot continue execution from another transaction')

        try:
            results = await self._exec_continue(limit=limit)
        finally:
            await self.conn._wait_for_ready()

        return results

    async def _exec_continue(self, limit=None):
        execute_msg = _pgmsg.Execute(self._portal_name, max_rows=limit)
        sync_msg = _pgmsg.Sync()
        await self.conn._send_msg(execute_msg, sync_msg)

        self.conn._is_ready = False

        results = []
        should_close = False
        while True:
            msg = await self.conn._get_msg(
                _pgmsg.ErrorResponse,
                _pgmsg.DataRow,
                _pgmsg.CommandComplete,
                _pgmsg.PortalSuspended)
            if isinstance(msg, _pgmsg.DataRow):
                row = self.conn._codec_helper.decode_row(
                    msg.columns, self._row_desc)
                results.append(row)
            elif isinstance(msg, _pgmsg.ErrorResponse):
                raise get_exc_from_msg(
                    msg,
                    desc_prefix=(
                        f'Error executing statement: {self.query}\n   '
                    ),
                )
            elif isinstance(msg, _pgmsg.CommandComplete):
                if msg.cmd_tag.startswith(b'SELECT'):
                    _, rows = msg.cmd_tag.split(b' ')
                    self._query_row_count = int(rows.decode('ascii'))
                else:
                    self._query_row_count = None
                should_close = True
                break
            elif isinstance(msg, _pgmsg.PortalSuspended):
                break
            else:
                assert False

        if should_close:
            await self.conn._wait_for_ready()
            await self._close_portal()

        return results

    async def _close_portal(self):
        if not self._execute_started or self._portal_closed:
            return

        close_msg = _pgmsg.Close(b'P', self._portal_name)
        sync_msg = _pgmsg.Sync()
        await self.conn._send_msg(close_msg, sync_msg)

        self.conn._is_ready = False

        # handle close portal response
        msg = await self.conn._get_msg(_pgmsg.CloseComplete,
                                       _pgmsg.ErrorResponse)
        if isinstance(msg, _pgmsg.ErrorResponse):
            raise get_exc_from_msg(
                msg,
                desc_prefix=(
                    f'Error closing portal: {self.query}\n   '
                ),
            )

        self._portal_closed = True

    async def forward(self, n):
        if self.conn.closed.is_set():
            raise ProgrammingError('Connection is closed.')

        if not self._execute_started:
            raise ProgrammingError(
                'exec_continue can only be called after execute is '
                'called')

        if not self.conn.in_transaction:
            # we don't allow this because outside a transaction
            # postgres will complain that "portal does not exist"
            raise InterfaceError(
                'Cannot continue execution of a statement outside '
                'a transaction')

        if self.conn.current_transaction != self._initial_transaction:
            raise ProgrammingError(
                'Cannot continue execution from another transaction')


        await self.conn.execute(
            f'move forward {n} from {self._portal_name}')

    def __repr__(self):
        return f'<PreparedStatement name={self._stmt_name} query="{self.query}">'

    def __del__(self):
        if hasattr(self, '_stmt_name') and self._stmt_name:
            self.conn._close_stmt(self._stmt_name)

    @property
    def finished(self):
        return self._portal_closed



import hashlib


#
# postgres message data types as defined here:
# https://www.postgresql.org/docs/current/protocol-message-types.html
#


class PgBaseDataType:
    pass


class _Int(int, PgBaseDataType):
    def __bytes__(self):
        return self.to_bytes(length=self._int_size,
                             byteorder='big',
                             signed=True)

    @classmethod
    def deserialize(cls, msg, start):
        value = msg[start:start+cls._int_size]
        value = int.from_bytes(value, byteorder='big', signed=True)
        return cls(value), cls._int_size


class Int8(_Int):
    _int_size = 1


class Int16(_Int):
    _int_size = 2


class Int32(_Int):
    _int_size = 4


class Byte1(bytes, PgBaseDataType):
    def __new__(cls, value):
        assert isinstance(value, bytes)
        assert len(value) == 1
        return super().__new__(cls, value)

    def __repr__(self):
        return f'<Byte1 {super().__repr__()} ({self[0]})>'

    @classmethod
    def deserialize(cls, msg, start):
        return cls(msg[start:start+1]), 1


class String(bytes, PgBaseDataType):
    def __new__(cls, value):
        if isinstance(value, str):
            value = value.encode('utf-8')
        assert isinstance(value, bytes)
        return super().__new__(cls, value)

    def __bytes__(self):
        return self + b'\0'

    def __repr__(self):
        return f'<String "{self}">'

    def __str__(self):
        try:
            value = self.decode('utf-8')
        except UnicodeDecodeError:
            value = super().__str__[1:]  # remove the b prefix
        return f'{value}'

    @classmethod
    def deserialize(cls, msg, start):
        try:
            null_idx = msg.index(b'\0', start)
        except ValueError:
            raise ValueError('String is not null terminated.')
        value = msg[start:null_idx]
        return cls(value), null_idx - start + 1


#
# postgres messages
#


# metaclass for all postgres message classes, applied through the base
# class PgMessage
class PgMessageMetaClass(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        if name != 'PgMessage':
            klass = PgMessageMetaClass._create_class(
                cls, name, bases, attrs, **kwargs)
        else:
            klass = super().__new__(cls, name, bases, attrs)

        return klass

    @staticmethod
    def _create_class(cls, name, bases, attrs, **kwargs):
        if '_type' not in attrs:
            # allow _type to be inherited
            for base_class in bases:
                if hasattr(base_class, '_type'):
                    # don't use the _type here though, because we've
                    # processed it already for the base class
                    _type = None
                    break
            else:
                raise TypeError(
                    f'Class {name} missing class variable "_type"')
        else:
            _type = attrs['_type']

        if 'side' not in kwargs:
            raise TypeError(
                f'Class {name} missing class keyword argument "side".')
        side = kwargs['side']

        if _type is not None:
            if not isinstance(_type, bytes) or len(_type) != 1:
                raise ValueError(
                    '_type field should contain a bytes object of '
                    'size 1 or None.')

        if side not in ['backend', 'frontend', 'both']:
            raise ValueError(
                'Class argument "side" can only have one of these '
                'values: backend, frontend, both')

        # store the side argument as a class variable
        attrs['_side'] = side

        for attr, value in attrs.items():
            if attr.startswith('_'):
                continue
            if not isinstance(value, PgBaseDataType) and \
               not (isinstance(value, type) and
                    issubclass(value, PgBaseDataType)) and \
               not callable(value):
                raise TypeError(
                    'PgMessage sub-class fields should either by a '
                    'sub-class of PgBaseDataType, or an instance of '
                    'such a sub-class, or a callable returning '
                    'bytes.')

        klass = super().__new__(cls, name, bases, attrs)
        if _type is not None and side != 'frontend':
            PgMessage._msg_classes[_type] = klass

        return klass


# this is the base class for all postgres messages. it adds a
# __bytes__ function that serializes an instance of a sub-class to
# bytes.
#
# see list of messages here:
# https://www.postgresql.org/docs/current/protocol-message-formats.html
class PgMessage(metaclass=PgMessageMetaClass):
    # maps message types to their relevant sub-class of
    # PgMessage. this will be populated by PgMessageMetaClass.
    _msg_classes = {}

    def __bytes__(self):
        # calculate payload
        klass = type(self)
        payload = bytearray()
        for attr, field in vars(klass).items():
            if attr.startswith('_'):
                continue
            if isinstance(field, PgBaseDataType):
                # the field contains a concrete value (like Int32(40),
                # which should always contain the value 40)
                value = bytes(field)
            elif isinstance(field, type) and \
                 issubclass(field, PgBaseDataType):
                # the field contains just a type (like Int32); the
                # field value should have been set in the object
                # itself before serialization attempt.
                value = getattr(self, attr)
                if not isinstance(value, PgBaseDataType):
                    # case it to the relevant PgBaseDataType type
                    value = field(value)
            elif callable(field):
                value = field(self)
            payload += bytes(value)

        # add four to length due to the length of the "length" field
        # itself
        length = 4 + len(payload)
        length = length.to_bytes(length=4, byteorder='big', signed=True)

        msg = self._type if self._type is not None else b''
        msg += length + payload

        return msg

    @classmethod
    def deserialize(cls, msg, start=0):
        if len(msg) - start < 5:
            # not enough data
            return None, 0

        msg_type = msg[start + 0:start + 1]
        subclass = PgMessage._msg_classes.get(msg_type)
        if subclass is None:
            raise ValueError(f'Unknown message type: {msg_type}')

        msg_len = msg[start + 1:start + 5]
        msg_len = int.from_bytes(msg_len, byteorder='big')

        if msg_len > len(msg) - start - 1:
            # not enough data
            return None, 0

        msg = subclass._deserialize(
            msg,
            start + 1 + 4,  # one byte for type, 4 for length
            msg_len - 4     # length consists of the length field
                            # itself but not type
        )

        # return the deserialized message, as well as the number of
        # bytes consumed.
        return msg, msg_len + 1

    @classmethod
    def _deserialize(cls, msg, start, length):
        # this is the default implementation of the _deserialize
        # method that can deserialize all messages, except the ones
        # with unusual structure. for those, the _deserialize class
        # method should be overridden in the relevant sub-class.

        msg_obj = cls()
        idx = start
        for attr, value in vars(cls).items():
            if attr.startswith('_'):
                continue

            if isinstance(value, PgBaseDataType):
                # the field contains a concrete value (like Int32(40),
                # which should always contain the value 40)
                field_type = type(value)
            elif isinstance(value, type) and \
                 issubclass(value, PgBaseDataType):
                # the field contains just a type (like Int32)
                field_type = value
            elif callable(value):
                raise ValueError(
                    'Cannot deserialize messages with dynamic fields.')
            else:
                # the metaclass validation should prevent this to ever
                # happen
                assert False

            field_value, field_len = field_type.deserialize(msg, idx)
            idx += field_len

            setattr(msg_obj, attr, field_value)

        return msg_obj


# this class handles the following message types: AuthenticationOk,
# AuthenticationKerberosV5, AuthenticationCleartextPassword,
# AuthenticationMD5Password, AuthenticationSCMCredential,
# AuthenticationGSS, AuthenticationGSSContinue, AuthenticationSSPI,
# AuthenticationSASL, AuthenticationSASLContinue, AuthenticationSASLFinal.
#
# all of these messages have the same _type field (R). the deserialize
# method returns the appropriate sub-class, depending on the value of
# the auth field.
class Authentication(PgMessage, side='backend'):
    _type = b'R'
    auth = Int32(0)
    # depending on the value of the auth field, more fields might
    # follow; a custom _deserialize function will handle these.

    @classmethod
    def _deserialize(cls, msg, start, length):
        auth, _ = Int32.deserialize(msg, start)
        if auth == 0:
            return AuthenticationOk()
        if auth == 2:
            return AuthenticationKerberosV5()
        if auth == 3:
            return AuthenticationCleartextPassword()
        if auth == 5:
            salt = msg[start+4:start+8]
            return AuthenticationMD5Password(salt)
        if auth == 6:
            return AuthenticationSCMCredential()
        if auth == 7:
            return AuthenticationGSS()
        if auth == 8:
            auth_data = msg[start+4:]
            return AuthenticationGSSContinue(auth_data)
        if auth == 9:
            return AuthenticationSSPI()
        if auth == 10:
            sasl_auth_mechanism, _ = String.deserialize(msg, start + 4)
            return AuthenticationSASL(sasl_auth_mechanism)
        if auth == 11:
            sasl_data = msg[start+4:]
            return AuthenticationSASLContinue(sasl_data)
        if auth == 12:
            sasl_additional_data = msg[start+4:]
            return AuthenticationSASLFinal(sasl_additional_data)
        raise ValueError(f'Unknown auth type: {auth}')


class AuthenticationOk(Authentication, side='backend'):
    auth = Int32(0)

    def __repr__(self):
        return '<AuthenticationOk>'


class AuthenticationKerberosV5(Authentication, side='backend'):
    auth = Int32(2)

    def __repr__(self):
        return '<AuthenticationKerberosV5>'


class AuthenticationCleartextPassword(Authentication, side='backend'):
    auth = Int32(3)

    def __repr__(self):
        return '<AuthenticationCleartextPassword>'


class AuthenticationMD5Password(Authentication, side='backend'):
    auth = Int32(5)

    def __init__(self, salt):
        self.salt = salt

    def __repr__(self):
        salt = repr(self.salt)[2:-1]  # string b prefix and quotes
        return f'<AuthenticationMD5Password salt="{salt}">'


class AuthenticationSCMCredential(Authentication, side='backend'):
    auth = Int32(6)

    def __repr__(self):
        return '<AuthenticationSCMCredential>'


class AuthenticationGSS(Authentication, side='backend'):
    auth = Int32(7)

    def __repr__(self):
        return '<AuthenticationGSS>'


class AuthenticationGSSContinue(Authentication, side='backend'):
    auth = Int32(8)

    def __init__(self, auth_data):
        self.auth_data = auth_data

    def __repr__(self):
        auth_data = repr(self.auth_data)[2:-1]  # string b prefix and
                                                # quotes
        return f'<AuthenticationGSSContinue auth_data={auth_data}>'


class AuthenticationSSPI(Authentication, side='backend'):
    auth = Int32(9)

    def __repr__(self):
        return '<AuthenticationSSPI>'


class AuthenticationSASL(Authentication, side='backend'):
    auth = Int32(10)

    def __init__(self, auth_mechanism):
        self.auth_mechanism = auth_mechanism

    def __repr__(self):
        auth_mech = repr(self.auth_mechanism)[2:-1]  # string b prefix
                                                     # and quotes
        return f'<AuthenticationSASL auth_mechanism={auth_mech}>'


class AuthenticationSASLContinue(Authentication, side='backend'):
    auth = Int32(11)

    def __init__(self, data):
        self.data = data

    def __repr__(self):
        data = repr(self.data)[2:-1]  # string b prefix and quotes
        return f'<AuthenticationSASLContinue auth_mechanism={data}>'


class AuthenticationSASLFinal(Authentication, side='backend'):
    auth = Int32(12)

    def __init__(self, data):
        self.data = data

    def __repr__(self):
        data = repr(self.data)[2:-1]  # string b prefix and quotes
        return f'<AuthenticationSASLFinal auth_mechanism={data}>'


class BackendKeyData(PgMessage, side='backend'):
    _type = b'K'
    pid = Int32
    secret_key = Int32

    def __repr__(self):
        return (
            f'<BackendKeyData pid={self.pid} '
            f'secret_key={self.secret_key}>'
        )


class Bind(PgMessage, side='frontend'):
    _type = b'B'
    portal_name = String
    stmt_name = String
    param_format_code_count = Int16
    # if param_format_code_count is non-zero, a number of Int16's
    # follow, specifying format codes of parameters (0 means text,
    # which is the default, and 1 means binary)
    param_count = Int16
    # if param_count is non-zero, a number of (Int32, ByteN) pairs
    # appear, specifying the length and content of parameters.
    result_format_code_count = Int16
    # if result_format_code_count is non-zero, a number of Int16's
    # appear here, specifying the format codes of each column of
    # results.

    def __init__(self, portal_name, stmt_name, params=None,
                 param_format_codes=None, result_format_codes=None):
        if not isinstance(stmt_name, (str, bytes)):
            raise TypeError('stmt_name should be a bytes or str')
        if not isinstance(portal_name, (str, bytes)):
            raise TypeError('portal_name should be a bytes or str')
        self.portal_name = portal_name
        self.stmt_name = stmt_name
        self.params = params or []
        self.param_format_codes = param_format_codes or []
        self.result_format_codes = result_format_codes or []

    def __bytes__(self):
        portal_name = bytes(String(self.portal_name))
        stmt_name = bytes(String(self.stmt_name))
        param_format_code_count = bytes(
            Int16(len(self.param_format_codes)))
        param_format_codes = b''.join(
            bytes(Int16(fc)) for fc in self.param_format_codes)
        param_count = bytes(Int16(len(self.params)))
        params = b''.join(
            # NULL values are indicated by setting the length field to
            # -1
            (
                bytes(Int32(len(p) if p is not None else -1)) +
                (b'' if p is None else p)
            )
            for p in self.params
        )
        result_format_code_count = bytes(
            Int16(len(self.result_format_codes)))
        result_format_codes = b''.join(
            bytes(Int16(fc)) for fc in self.result_format_codes)
        length = bytes(Int32(
            4 +
            len(portal_name) +
            len(stmt_name) +
            len(param_format_code_count) +
            len(param_format_codes) +
            len(param_count) +
            len(params) +
            len(result_format_code_count) +
            len(result_format_codes)
        ))
        return (
            b'B' +
            length +
            portal_name +
            stmt_name +
            param_format_code_count +
            param_format_codes +
            param_count +
            params +
            result_format_code_count +
            result_format_codes
        )

    def __repr__(self):
        params = ''
        if self.params:
            params = f' params={self.params!r}'
        pcodes = ''
        if self.param_format_codes:
            pcodes = f' param_fmt={[int(c) for c in self.param_format_codes]!r}'
        rcodes = ''
        if self.result_format_codes:
            rcodes = f' result_fmt={[int(c) for c in self.result_format_codes]!r}'
        return (
            f'<Bind portal={self.portal_name} stmt={self.stmt_name}'
            f'{params}{pcodes}{rcodes}>'
        )


class BindComplete(PgMessage, side='backend'):
    _type = b'2'

    def __repr__(self):
        return '<BindComplete>'


class CommandComplete(PgMessage, side='backend'):
    _type = b'C'
    cmd_tag = String

    def __repr__(self):
        return f'<CommandComplete tag="{self.cmd_tag}">'


class Close(PgMessage, side='frontend'):
    _type = b'C'
    type = Byte1
    name = String

    def __init__(self, stmt_or_portal, name):
        assert stmt_or_portal in (b'S', b'P')
        self.type = stmt_or_portal

        if isinstance(name, str):
            name = name.encode('utf-8')
        self.name = name

    def __bytes__(self):
        type_bytes = bytes(self.type)
        name_bytes = bytes(String(self.name))
        length = bytes(Int32(4 + len(type_bytes) + len(name_bytes)))
        return b'C' + length + type_bytes + name_bytes

    def __repr__(self):
        type = self.type.decode('ascii')
        try:
            name = self.name.decode('ascii')
        except UnicodeDecodeError:
            name = repr(self.name)[2:-1]
        return f'<Close type={type} name={name}>'


class CloseComplete(PgMessage, side='backend'):
    _type = b'3'

    def __repr__(self):
        return '<CloseComplete>'


class DataRow(PgMessage, side='backend'):
    _type = b'D'
    column_count = Int16

    # for each column there will be an Int32, telling us the length of
    # the column value, and a ByteN containing the value.

    def __init__(self, columns=None):
        self.columns = columns
        if self.columns is None:
            self.columns = []

    def __repr__(self):
        columns = ' '.join(repr(c) for c in self.columns)
        return f'<DataRow {columns}>'

    @classmethod
    def _deserialize(cls, msg, start, length):
        obj = cls()

        idx = start
        obj.column_count, n = Int16.deserialize(msg, idx)
        idx += n

        for i in range(obj.column_count):
            column_size, n = Int32.deserialize(msg, idx)
            idx += n

            if column_size < 0:
                # this is a NULL value
                column_value = None
            else:
                column_value = msg[idx:idx+column_size]
                idx += column_size

            obj.columns.append(column_value)

        return obj


class Describe(PgMessage, side='frontend'):
    _type = b'D'
    type = Byte1
    name = String

    def __init__(self, stmt_or_portal, name):
        assert stmt_or_portal in (b'S', b'P')
        self.type = stmt_or_portal

        if isinstance(name, str):
            name = name.encode('utf-8')
        self.name = name

    def __bytes__(self):
        type_bytes = bytes(self.type)
        name_bytes = bytes(String(self.name))
        length = bytes(Int32(4 + len(type_bytes) + len(name_bytes)))
        return b'D' + length + type_bytes + name_bytes

    def __repr__(self):
        type = self.type.decode('ascii')
        try:
            name = self.name.decode('ascii')
        except UnicodeDecodeError:
            name = repr(self.name)[2:-1]
        return f'<Describe type={type} name={name}>'


class EmptyQueryResponse(PgMessage, side='backend'):
    _type = b'I'

    def __repr__(self):
        return '<EmptyQueryResponse>'


class ErrorResponse(PgMessage, side='backend'):
    _type = b'E'
    # fields consist of one or more of (Byte1, String) pairs

    def __init__(self, pairs):
        self.pairs = pairs

    def __repr__(self):
        nfields = f'({len(self.pairs)} field(s))'
        pairs = ' '.join(f'{code}={value}' for code, value in self.pairs)
        return f'<ErrorResponse {nfields} {pairs}>'

    @classmethod
    def _deserialize(cls, msg, start, length):
        pairs = []
        idx = start
        while True:
            if idx >= len(msg):
                raise ValueError(
                    'Unterminated ErrorResponse message (should end '
                    'with a zero byte)')
            code, n = Byte1.deserialize(msg, idx)
            if code == b'\0':
                break
            code = chr(code[0])  # convert e.g. b'C' to 'C'
            idx += n
            value, n = String.deserialize(msg, idx)
            idx += n
            pairs.append((code, value))
        obj = cls(pairs)
        return obj


class Execute(PgMessage, side='frontend'):
    _type = b'E'
    portal_name = String
    max_rows = Int32

    def __init__(self, portal_name, max_rows=None):
        self.portal_name = portal_name
        self.max_rows = max_rows or 0

    def __bytes__(self):
        portal_name = bytes(String(self.portal_name))
        max_rows = bytes(Int32(self.max_rows))
        length = bytes(Int32(4 + len(portal_name) + len(max_rows)))
        return b'E' + length + portal_name + max_rows

    def __repr__(self):
        max_rows = ''
        if self.max_rows:
            max_rows = f' max_rows={self.max_rows}'
        return f'<Execute portal={self.portal_name}{max_rows}>'


class Flush(PgMessage, side='frontend'):
    _type = b'H'


class NegotiateProtocolVersion(PgMessage, side='backend'):
    _type = b'v'
    minor_ver_supported = Int32
    n_unrecognized_proto_options = Int32
    option_name = String


class NoData(PgMessage, side='backend'):
    _type = b'n'

    def __repr__(self):
        return '<NoData>'


class NoticeResponse(PgMessage, side='backend'):
    _type = b'N'
    # the message body contains one or more of (Byte1, String) pairs
    # each denoting a code and a value. The custom _deserializer
    # gathers these in a field named "notices".

    def __init__(self, pairs=None):
        self.pairs = pairs or []

    def __repr__(self):
        nfields = f'({len(self.notices)} field(s))'
        notices = ' '.join(
            f'{code}={value}' for code, value in self.notices)
        return f'<NoticeMessage {nfields} {notices}>'

    @classmethod
    def _deserialize(cls, msg, start, length):
        pairs = []
        idx = start
        while True:
            if idx >= len(msg):
                raise ValueError(
                    'Unterminated NoticeMessage message (should end '
                    'with a zero byte)')
            code, n = Byte1.deserialize(msg, idx)
            if code == b'\0':
                break
            code = chr(code[0])  # convert e.g. b'C' to 'C'
            idx += n
            value, n = String.deserialize(msg, idx)
            idx += n
            pairs.append((code, value))
        obj = cls(pairs)
        return obj


class ParameterDescription(PgMessage, side='backend'):
    _type = b't'
    nparams = Int16
    # after this, for each parameter an oid value follows. these will
    # be gather in the param_oids field.

    def __init__(self, param_oids=[]):
        self.param_oids = []

    def __repr__(self):
        return f'<ParameterDescription {self.param_oids}>'

    @classmethod
    def _deserialize(cls, msg, start, length):
        obj = cls()
        idx = start
        obj.nparams, n = Int16.deserialize(msg, idx)
        idx += n
        while len(obj.param_oids) < obj.nparams:
            oid, n = Int32.deserialize(msg, idx)
            idx += n
            obj.param_oids.append(oid)
        return obj


class ParameterStatus(PgMessage, side='backend'):
    _type = b'S'
    param_name = String
    param_value = String

    def __repr__(self):
        return (
            f'<ParameterStatus name={self.param_name} '
            f'value={self.param_value}>'
        )


class Parse(PgMessage, side='frontend'):
    _type = b'P'
    name = String
    query = String
    param_type_count = Int16

    # if param_type_count is non-zero, a number of Int32's follow,
    # specifying the type of parameters

    def __init__(self, name, query, param_types=None):
        if isinstance(name, str):
            name = name.encode('utf-8')
        if isinstance(query, str):
            query = query.encode('utf-8')
        self.name = name
        self.query = query
        self.param_types = param_types or []

    def __bytes__(self):
        name = bytes(String(self.name))
        query = bytes(String(self.query))
        param_types = b''.join(
            bytes(Int32(pt)) for pt in self.param_types)
        param_type_count = bytes(Int16(len(self.param_types)))
        length = bytes(Int32(
            4 +
            len(name) +
            len(query) +
            len(param_type_count) +
            len(param_types)
        ))
        return (
            b'P' +
            length +
            name +
            query +
            param_type_count +
            param_types
        )

    def __repr__(self):
        name = self.name
        if isinstance(name, bytes):
            name = name.decode('utf-8')
        query = self.query
        if isinstance(query, bytes):
            query = query.decode('utf-8')
        return (
            f'<Parse name={name} query="{query}" '
            f'param_types={self.param_types}>'
        )


class ParseComplete(PgMessage, side='backend'):
    _type = b'1'

    def __repr__(self):
        return f'<ParseComplete>'


class PasswordMessage(PgMessage, side='frontend'):
    _type = b'p'
    password = String

    def __init__(self, password, *, md5=False, username=None,
                 salt=None):
        if isinstance(password, str):
            password = password.encode('utf-8')

        if md5:
            if salt is None:
                raise ValueError('salt is required for MD5 password')

            username = username or b''
            if isinstance(username, str):
                username = username.encode('utf-8')

            # calculate md5 hash as described here:
            # https://www.postgresql.org/docs/current/auth-password.html
            userpass = password + username
            userpass_md5 = (
                hashlib.md5(userpass)
                .hexdigest()
                .encode('ascii')
            )
            final_hash = (
                hashlib.md5(userpass_md5 + salt)
                .hexdigest()
                .encode('ascii')
            )
            password = b'md5' + final_hash

        self.password = password

    def __repr__(self):
        password = repr(bytes(self.password))[2:-1]
        return f'<PasswordMessage password="{password}">'


class PortalSuspended(PgMessage, side='backend'):
    _type = b's'


class Query(PgMessage, side='frontend'):
    _type = b'Q'
    query = String

    def __init__(self, query):
        if isinstance(query, str):
            query = query.encode('utf-8')
        if not isinstance(query, bytes):
            raise ValueError(
                'query should either by a str or bytes; got '
                f'{type(query)}')

        self.query = query

    def __repr__(self):
        query = self.query
        if isinstance(query, bytes):
            query = query.decode('utf-8')
        return f'<Query {query}>'


class ReadyForQuery(PgMessage, side='backend'):
    _type = b'Z'
    status = Byte1

    def __repr__(self):
        try:
            status = self.status.decode('utf-8')
        except UnicodeDecodeError:
            status = str(self.status)

        if status == 'I':
            status_desc = 'idle'
        elif status == 'T':
            status_desc = 'inside-transaction-block'
        elif status == 'E':
            status_desc = 'error'
        else:
            status_desc = 'unknown'
        status = f'{status} ({status_desc})'
        return f'<ReadyForQuery status="{status}">'


class RowDescription(PgMessage, side='backend'):
    _type = b'T'
    nfields = Int16
    # after this, for each row field a number of fields follow, parsed
    # in the custom _deserialize method

    def __init__(self, fields=[]):
        self.fields = []

    def __repr__(self):
        return f'<RowDescription {self.fields}>'

    @classmethod
    def _deserialize(cls, msg, start, length):
        obj = cls()
        idx = start
        obj.nfields, n = Int16.deserialize(msg, idx)
        idx += n
        while len(obj.fields) < obj.nfields:
            name, n = String.deserialize(msg, idx)
            idx += n
            table_oid, n = Int32.deserialize(msg, idx)
            idx += n
            column_attr_id, n = Int16.deserialize(msg, idx)
            idx += n
            type_oid, n = Int32.deserialize(msg, idx)
            idx += n
            type_len, n = Int16.deserialize(msg, idx)
            idx += n
            type_modifier, n = Int32.deserialize(msg, idx)
            idx += n
            format_code, n = Int16.deserialize(msg, idx)
            idx += n
            obj.fields.append((name, table_oid, column_attr_id,
                               type_oid, type_len, type_modifier,
                               format_code))
        return obj


class SSLRequest(PgMessage, side='frontend'):
    _type = None  # SSLRequest message has no type field
    ssl_request_code = Int32(80877103)


class StartupMessage(PgMessage, side='frontend'):
    _type = None  # startup message has no type field
    version = Int32(0x0003_0000)  # protocol version 3.0
    # params = dynamic field

    def __init__(self, user, database):
        self.user = user.encode('utf-8')
        self.database = database.encode('utf-8')

    def __repr__(self):
        user = self.user.decode('utf-8')
        database = self.database.decode('utf-8')
        return f'<StartupMessage user={user} db={database}>'

    def params(self):
        return (
            b'user\0' +
            self.user + b'\0' +
            b'database\0' +
            self.database + b'\0' +
            b'datestyle\0iso\0' +
            b'intervalstyle\0iso_8601\0' +
            b'\0'
        )


class Sync(PgMessage, side='frontend'):
    _type = b'S'

    def __repr__(self):
        return '<Sync>'


class Terminate(PgMessage, side='frontend'):
    _type = b'X'

    def __repr__(self):
        return '<Terminate>'



class Warning(Exception):
    pass


class Error(Exception):
    def __init__(self, error_msg):
        self.error_msg = error_msg


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    def __init__(self, error_msg, severity=None):
        self.severity = severity
        self.error_msg = error_msg

    def __str__(self):
        severity = self.severity or 'UNKNOWN-SEVERITY'
        return f'Database error: [{severity}] {self.error_msg}'


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass



import warnings
import struct
import socket
import re
import dateutil.parser
import orjson
from datetime import date, time, datetime, timedelta, timezone
from ipaddress import (
    ip_address, ip_network, IPv4Address, IPv6Address, IPv4Network,
    IPv6Network,
)
from functools import wraps
from ._exceptions import InterfaceError
from ._utils import PgProtocolFormat

interval_re = re.compile(
    r'^(P((?P<years>[-\.\d]+?)Y)?((?P<months>[-\.\d]+?)M)?((?P<weeks>[-\.\d]+?)W)?((?P<days>[-\.\d]+?)D)?)?'
    r'(T((?P<hours>[-\.\d]+?)H)?((?P<minutes>[-\.\d]+?)M)?((?P<seconds>[-\.\d]+?)S)?)?$'
)
pg_epoch = datetime(2000, 1, 1)
pg_epoch_utc = datetime(2000, 1, 1, tzinfo=timezone.utc)

builtin_codecs = {}


def register_builtin_codec(codec):
    builtin_codecs[codec.pg_type] = codec
    return codec


class CodecHelper:
    def __init__(self):
        self.initialized = False

        self._codecs = builtin_codecs
        self._oid_to_name = {}
        self._type_name_to_codec = {}

    def init(self, pg_types):
        for name, oid in pg_types:
            # values passed to this message are un-decoded and in text
            # format (because we haven't initialized codecs yet).
            name = name.decode('ascii')
            oid = int(oid.decode('ascii'))
            if name in self._codecs:
                self.enable_codec(name, oid)
        self.initialized = True

    def enable_codec(self, type_name, type_oid):
        self._oid_to_name[type_oid] = type_name

        for name, codec in self._codecs.items():
            if name == type_name:
                self._type_name_to_codec[name] = codec
                break

    def register_codec(self, codec):
        if not isinstance(codec, type) or not issubclass(codec, Codec):
            raise TypeError(
                'Codec should be a sub-class of Codec class')
        self._codecs[codec.pg_type] = codec
        self._type_name_to_codec[codec.pg_type] = codec

    def encode_value(self, value, type_oid, protocol_format):
        if value is None:
            return None

        type_name = self._oid_to_name.get(type_oid)
        if type_name is None:
            raise InterfaceError(f'Uknown type OID: {type_oid}')

        codec = self._type_name_to_codec.get(type_name)
        if type_name is None:
            raise InterfaceError(
                f'No codec to encode parameter "{value}" to type '
                f'"{type_name}"')

        if protocol_format == PgProtocolFormat.TEXT:
            return codec.encode_text(value)
        else:
            return codec.encode_binary(value)

    def decode_row(self, columns, row_desc):
        row = []
        for col, col_desc in zip(columns, row_desc):
            row.append(self.decode_col(col, col_desc))
        return tuple(row)

    def decode_col(self, col, col_desc):
        if col is None:
            return None

        (
            name, table_oid, column_attr_id, type_oid, type_len,
            type_modifier, format_code
        ) = col_desc

        type_name = self._oid_to_name.get(type_oid)
        if type_name is None:
            warnings.warn(
                f'Unknown column OID: {type_oid}; returning un-decoded '
                'data.')
            return col

        codec = self._type_name_to_codec.get(type_name)
        if codec is None:
            warnings.warn(
                f'Unknown column type: {type_name}; returning '
                'un-decoded data.')
            return col

        if format_code == 0:
            return codec.decode_text(col)
        elif format_code == 1:
            return codec.decode_binary(col)
        else:
            raise InterfaceError(
                f'Unexpected column format code: {format_code}')


class CodecMetaclass(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        klass = super().__new__(cls, name, bases, attrs)
        if name == 'Codec' or name.startswith('_'):
            return klass

        if not hasattr(klass, 'pg_type'):
            raise TypeError('Codec class does not have a pg_type field')

        if not isinstance(klass.pg_type, str):
            raise TypeError(
                f'pg_type field must be a string; Got: '
                f'{klass.pg_type!r}')

        if hasattr(klass, 'python_types'):
            if isinstance(klass.python_types, type):
                # convert single values to tuples
                klass.python_types = (klass.python_types,)
            if isinstance(klass.python_types, list):
                klass.python_types = tuple(klass.python_types)
            if not isinstance(klass.python_types, tuple):
                raise TypeError(
                    f'Codec class python_types field should either be '
                    f'a tuple or a type; Got "{klass.python_types}"')

        expected_methods = [
            'decode_text',
            'decode_binary',
            'encode_text',
            'encode_binary',
        ]
        for method_name in expected_methods:
            if not hasattr(klass, method_name):
                raise TypeError(
                    f'Method {method_name} not found in codec class '
                    f'{name}')
            from inspect import getattr_static
            method = getattr_static(klass, method_name)
            if not isinstance(method, classmethod) and \
               not isinstance(method, staticmethod):
                raise TypeError(
                    'Codec encode/decode methods should either be '
                    'class methods or static methods.')

            # add type checkers to encode methods
            if klass.python_types:
                setattr(klass, method_name,
                        cls.get_typechecked_method(
                            klass,
                            getattr(klass, method_name),
                            klass.python_types,
                            klass.pg_type))

        return klass

    @classmethod
    def get_typechecked_method(cls, codec_class, method, types,
                               pg_type_name):
        @wraps(method)
        def encode_wrapper(value, *args, **kwargs):
            if not isinstance(value, types):
                if len(types) == 1:
                    raise TypeError(
                        f'Codec {codec_class.__name__} expects a value '
                        f'of python type "{types[0].__name__}" for '
                        f'postgres type "{pg_type_name}"; Got: '
                        f'{value!r}')
                else:
                    types_str = f'[{", ".join(t.__name__ for t in types)}]'
                    raise TypeError(
                        f'Codec {codec_class.__name__} expects a value '
                        f'of a python type in {types_str} for postgres '
                        f'type "{pg_type_name}"; Got: {value!r}')

            result = method(value, *args, **kwargs)
            if not isinstance(result, bytes):
                raise TypeError(
                    f'Codec {codec_class.__name__} {method.__name__} '
                    f'method did not return a bytes value; Got: '
                    f'{result!r}')
            return result

        @wraps(method)
        def decode_wrapper(value, *args, **kwargs):
            if not isinstance(value, bytes):
                raise TypeError(
                    f'{method.__name__} expects a bytes value')
            result = method(value, *args, **kwargs)
            if method.__name__.startswith('decode_'):
                if not isinstance(result, types):
                    expected_types = \
                        f'[{", ".join(t.__name__ for t in types)}]'
                    raise TypeError(
                        f'Codec {codec_class.__name__} returned an '
                        f'invalid decoded value; Expected types: '
                        f'{expected_types}, Got: {result!r}')
            return result

        if method.__name__.startswith('decode_'):
            return decode_wrapper
        else:
            return encode_wrapper


class Codec(metaclass=CodecMetaclass):
    pass


class _Int(Codec):
    @classmethod
    def decode_text(cls, value):
        return int(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        return int.from_bytes(value, byteorder='big', signed=True)

    @classmethod
    def encode_text(cls, value):
        return str(value).encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        return value.to_bytes(length=cls._int_length,
                              byteorder='big',
                              signed=True)


@register_builtin_codec
class Int2(_Int):
    pg_type = 'int2'
    python_types = [int]
    _int_length = 2


@register_builtin_codec
class Int4(_Int):
    pg_type = 'int4'
    python_types = [int]
    _int_length = 4


@register_builtin_codec
class Int8(_Int):
    pg_type = 'int8'
    python_types = [int]
    _int_length = 8


@register_builtin_codec
class Bool(Codec):
    pg_type = 'bool'
    python_types = [bool]

    @classmethod
    def decode_text(cls, value):
        return value == b't'

    @classmethod
    def decode_binary(cls, value):
        return value[0] == 1

    @classmethod
    def encode_text(cls, value):
        return b't' if value else b'f'

    @classmethod
    def encode_binary(cls, value):
        return b'\x01' if value else b'\x00'


@register_builtin_codec
class ByteA(Codec):
    pg_type = 'bytea'
    python_types = [bytes]

    @classmethod
    def decode_text(cls, value):
        assert value.startswith(b'\\x')
        return bytes.fromhex(value[2:].decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        return value

    @classmethod
    def encode_text(cls, value):
        return b'\\x' + value.hex().encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        return value


@register_builtin_codec
class Text(Codec):
    pg_type = 'text'
    python_types = [str]

    @classmethod
    def decode_text(cls, value):
        return value.decode('utf-8')

    @classmethod
    def decode_binary(cls, value):
        return value.decode('utf-8')

    @classmethod
    def encode_text(cls, value):
        return value.encode('utf-8')

    @classmethod
    def encode_binary(cls, value):
        return value.encode('utf-8')


@register_builtin_codec
class Json(Codec):
    pg_type = 'json'
    python_types = [dict, list]

    @classmethod
    def decode_text(cls, value):
        return orjson.loads(value)

    @classmethod
    def decode_binary(cls, value):
        return orjson.loads(value)

    @classmethod
    def encode_text(cls, value):
        return orjson.dumps(value)

    @classmethod
    def encode_binary(cls, value):
        return orjson.dumps(value)


@register_builtin_codec
class Jsonb(Codec):
    pg_type = 'jsonb'
    python_types = [dict, list]

    @classmethod
    def decode_text(cls, value):
        return orjson.loads(value)

    @classmethod
    def decode_binary(cls, value):
        # create memory view so that we can slice without copying
        m = memoryview(value)
        version = m[0]
        if version != 1:
            raise InterfaceError(f'Unexpected JSONB format: {version}')
        return orjson.loads(m[1:])

    @classmethod
    def encode_text(cls, value):
        return orjson.dumps(value)

    @classmethod
    def encode_binary(cls, value):
        return b'\x01' + orjson.dumps(value)


@register_builtin_codec
class Float4(Codec):
    pg_type = 'float4'
    python_types = [float, int]

    @classmethod
    def decode_text(cls, value):
        return float(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        value, = struct.unpack('!f', value)
        return value

    @classmethod
    def encode_text(cls, value):
        return str(value).encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        return struct.pack('!f', value)


@register_builtin_codec
class Float8(Codec):
    pg_type = 'float8'
    python_types = [float, int]

    @classmethod
    def decode_text(cls, value):
        return float(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        value, = struct.unpack('!d', value)
        return value

    @classmethod
    def encode_text(cls, value):
        return str(value).encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        return struct.pack('!d', value)


@register_builtin_codec
class Inet(Codec):
    pg_type = 'inet'
    python_types = [IPv4Address, IPv6Address]

    @classmethod
    def decode_text(cls, value):
        return ip_address(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        return decode_inet_or_cidr(value)

    @classmethod
    def encode_text(cls, value):
        return str(value).encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        return encode_inet_or_cidr(value)


@register_builtin_codec
class Cidr(Codec):
    pg_type = 'cidr'
    python_types = [IPv4Network, IPv6Network]

    @classmethod
    def decode_text(cls, value):
        return ip_network(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        return decode_inet_or_cidr(value)

    @classmethod
    def encode_text(cls, value):
        return str(value).encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        return encode_inet_or_cidr(value)


@register_builtin_codec
class Char(Codec):
    pg_type = 'bpchar'
    python_types = [str]

    @classmethod
    def decode_text(cls, value):
        return value.decode('utf-8')

    @classmethod
    def decode_binary(cls, value):
        return value.decode('utf-8')

    @classmethod
    def encode_text(cls, value):
        return value[0].encode('utf-8')

    @classmethod
    def encode_binary(cls, value):
        return value[0].encode('utf-8')


@register_builtin_codec
class Varchar(Codec):
    pg_type = 'varchar'
    python_types = [str]

    @classmethod
    def decode_text(cls, value):
        return value.decode('utf-8')

    @classmethod
    def decode_binary(cls, value):
        return value.decode('utf-8')

    @classmethod
    def encode_text(cls, value):
        return value.encode('utf-8')

    @classmethod
    def encode_binary(cls, value):
        return value.encode('utf-8')


@register_builtin_codec
class Date(Codec):
    pg_type = 'date'
    python_types = [date]

    @classmethod
    def decode_text(cls, value):
        return dateutil.parser.parse(value.decode('ascii')).date()

    @classmethod
    def decode_binary(cls, value):
        value = int.from_bytes(value, byteorder='big', signed=True)
        return (pg_epoch + timedelta(days=value)).date()

    @classmethod
    def encode_text(cls, value):
        return value.isoformat().encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        # turn it into a datetime so we can subtract epoch from it
        value = datetime.combine(value, time(0, 0, 0))

        days = (value - pg_epoch).days
        return days.to_bytes(length=4, byteorder='big', signed=True)


@register_builtin_codec
class Time(Codec):
    pg_type = 'time'
    python_types = [time]

    @classmethod
    def decode_text(cls, value):
        return dateutil.parser.parse(value.decode('ascii')).time()

    @classmethod
    def decode_binary(cls, value):
        value = int.from_bytes(value, byteorder='big', signed=True)

        # the epoch value doesn't really matter here, since we strip
        # off the date part
        return (pg_epoch + timedelta(microseconds=value)).time()

    @classmethod
    def encode_text(cls, value):
        return value.isoformat().encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        useconds = (
            value.hour * 3600 * 1_000_000 +
            value.minute * 60 * 1_000_000 +
            value.second * 1_000_000 +
            value.microsecond
        )
        return useconds.to_bytes(length=8, byteorder='big', signed=True)


@register_builtin_codec
class DateTime(Codec):
    pg_type = 'timestamp'
    python_types = [datetime]

    @classmethod
    def decode_text(cls, value):
        return dateutil.parser.isoparse(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        value = int.from_bytes(value, byteorder='big', signed=True)
        return pg_epoch + timedelta(microseconds=value)

    @classmethod
    def encode_text(cls, value):
        return value.isoformat().encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        useconds = int((value - pg_epoch).total_seconds() * 1_000_000)
        return useconds.to_bytes(length=8, byteorder='big', signed=True)


@register_builtin_codec
class DateTimeTz(Codec):
    pg_type = 'timestamptz'
    python_types = [datetime]

    @classmethod
    def decode_text(cls, value):
        return dateutil.parser.isoparse(value.decode('ascii'))

    @classmethod
    def decode_binary(cls, value):
        value = int.from_bytes(value, byteorder='big', signed=True)
        dt = timedelta(microseconds=value)
        return (pg_epoch_utc + dt).astimezone()

    @classmethod
    def encode_text(cls, value):
        return value.isoformat().encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        seconds = (value - pg_epoch_utc).total_seconds()
        useconds = int(seconds * 1_000_000)
        return useconds.to_bytes(length=8, byteorder='big', signed=True)


@register_builtin_codec
class Interval(Codec):
    pg_type = 'interval'
    python_types = [timedelta]

    @classmethod
    def decode_text(cls, value):
        value = value.decode('ascii')

        m = interval_re.match(value)
        groups = m.groupdict()

        values = {}
        for i in ['years', 'months', 'weeks', 'days', 'hours',
                  'minutes', 'seconds']:
            values[i] = float(groups.get(i) or 0)

        days = values['years'] * 365 + values['months'] * 30 + values['days']
        del values['years']
        del values['months']
        del values['days']

        return timedelta(days=days, **values)

    @classmethod
    def decode_binary(cls, value):
        # see the interval_recv function in
        # src/backend/utils/adt/timestamp.c

        time = value[:8]
        days = value[8:12]
        months = value[12:]

        time = int.from_bytes(time, byteorder='big', signed=True)
        days = int.from_bytes(days, byteorder='big', signed=True)
        months = int.from_bytes(months, byteorder='big', signed=True)

        if months < 0:
            years = -(-months // 12)
            months = -(-months % 12)
        else:
            years = months // 12
            months = months % 12

        time = timedelta(microseconds=time)
        time += timedelta(days=days + months * 30 + years * 365)

        return time

    @classmethod
    def encode_text(cls, value):
        # adapted from: https://github.com/RusticiSoftware/TinCanPython
        # (Apache License 2.0)

        seconds = value.total_seconds()
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        days, hours = divmod(hours, 24)
        days, hours, minutes = map(int, (days, hours, minutes))
        seconds = round(seconds, 6)

        ## build date
        date = ''
        if days:
            date = f'{days}D'

        ## build time
        time = 'T'
        # hours
        bigger_exists = date or hours
        if bigger_exists and hours:
            time += f'{hours:02}H'
        # minutes
        bigger_exists = bigger_exists or minutes
        if bigger_exists and minutes:
          time += f'{minutes:02}M'
        # seconds
        if seconds.is_integer():
            seconds = f'{int(seconds):02}'
        else:
            # 9 chars long w/leading 0, 6 digits after decimal
            seconds = f'{seconds:09.6f}'
        # remove trailing zeros
        seconds = seconds.rstrip('0')
        if seconds:
            time += f'{seconds}S'

        iso_duration = 'P' + date + time
        return iso_duration.encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        useconds = value.microseconds
        useconds += value.seconds * 1_000_000

        # try not to use months if we can (because we'd lose accuracy
        # when we assume a month is always 30 days)
        days = value.days
        if days >= 2**32:
            months = days // 30
            days -= months * 30
        else:
            months = 0

        time = useconds.to_bytes(length=8, byteorder='big', signed=True)
        days = days.to_bytes(length=4, byteorder='big', signed=True)
        months = months.to_bytes(length=4, byteorder='big', signed=True)

        return time + days + months


@register_builtin_codec
class TimeTz(Codec):
    pg_type = 'timetz'
    python_types = [time]

    @classmethod
    def decode_text(cls, value):
        return dateutil.parser.parse(value.decode('ascii')).timetz()

    @classmethod
    def decode_binary(cls, value):
        # see the timetz_recv function in src/backend/utils/adt/date.c
        # in postgresql source code

        timeval = value[:8]
        zone = value[8:]

        useconds = int.from_bytes(timeval, byteorder='big', signed=True)
        zone = int.from_bytes(zone, byteorder='big', signed=True)

        hours = useconds // 3_600_000_000
        useconds -= hours * 3_600_000_000
        minutes = useconds // 60_000_000
        useconds -= minutes * 60_000_000
        seconds = useconds // 1_000_000
        useconds -= seconds * 1_000_000

        # Finally found out the reason for the following line in asyncpg
        # source code (in asyncpg/pgproto/codecs/datetime.pyz, in the
        # timetz_encode function [don't forget to get git submodules in
        # the asyncpg repo to get that file]): "In Python utcoffset() is
        # the difference between the local time and the UTC, whereas in
        # PostgreSQL it's the opposite, so we need to flip the sign."
        zone = -zone

        zone = timezone(timedelta(seconds=zone))

        return time(hours, minutes, seconds, useconds, tzinfo=zone)

    @classmethod
    def encode_text(cls, value):
        return value.isoformat().encode('ascii')

    @classmethod
    def encode_binary(cls, value):
        # see the timetz_send function in src/backend/utils/adt/date.c
        # in postgresql source code

        if value.utcoffset() is None:
            raise ValueError('No tzinfo in the passed time object')

        # for negation reason see decode_binary method
        zone = -int(value.utcoffset().total_seconds())

        useconds = (
            value.hour * 3_600_000_000 +
            value.minute * 60_000_000 +
            value.second * 1_000_000 +
            value.microsecond
        )

        zone = zone.to_bytes(length=4,
                             byteorder='big',
                             signed=True)
        useconds = useconds.to_bytes(length=8,
                                     byteorder='big',
                                     signed=True)

        return useconds + zone


## utils

def decode_inet_or_cidr(value):
    # see thet network_recv function in
    # src/backend/utils/adt/network.c in postgresql source code
    #
    # from the function docs:
    #
    # The external representation is (one byte apiece for) family,
    # bits, is_cidr, address length, address in network byte
    # order. Presence of is_cidr is largely for historical
    # reasons...We send it correctly on output, but ignore the value
    # on input.

    family = value[0]
    bits = value[1]
    is_cidr = value[2]
    address_length = value[3]
    address = value[4:]

    # these are defined in src/include/utils/inet.h. notice that
    # PGSQL_AF_INET6 value is not the same as socket.AF_INET6
    PGSQL_AF_INET = socket.AF_INET + 0
    PGSQL_AF_INET6 = socket.AF_INET + 1

    if family == PGSQL_AF_INET:
        address_class = IPv4Address
        network_class = IPv4Network
        max_prefixlen = 32
        ip_version = 4
    elif family == PGSQL_AF_INET6:
        address_class = IPv6Address
        network_class = IPv6Network
        max_prefixlen = 128
        ip_version = 6
    else:
        raise InterfaceError(
            'Invalid address family code received: {family}')

    if bits > max_prefixlen:
        raise InterfaceError(
            f'Invalid IPv{ip_version} CIDR netmask received: /{bits}')

    if len(address) != address_length:
        raise InterfaceError(
            'Invalid network address: address_length does not match '
            'length of data')

    if bits == max_prefixlen:
        return address_class(address)
    else:
        return network_class((address, bits))


def encode_inet_or_cidr(value):
    # see decode_binary comments

    # these are defined in src/include/utils/inet.h. notice that
    # PGSQL_AF_INET6 value is not the same as socket.AF_INET6
    PGSQL_AF_INET = socket.AF_INET + 0
    PGSQL_AF_INET6 = socket.AF_INET + 1

    if isinstance(value, (IPv4Address, IPv4Network)):
        family = PGSQL_AF_INET
    else:
        family = PGSQL_AF_INET6

    if isinstance(value, IPv4Address):
        bits = 32
        address_length = 4
    elif isinstance(value, IPv6Address):
        bits = 128
        address_length = 16
    elif isinstance(value, IPv4Network):
        bits = value.prefixlen
        address_length = 4
    else:
        bits = value.prefixlen
        address_length = 16

    is_cidr = 1 if isinstance(value, (IPv4Network, IPv6Network)) else 0
    header = struct.pack('!BBBB', family, bits, is_cidr, address_length)

    if isinstance(value, (IPv4Network, IPv6Network)):
        return header + value.network_address.packed
    else:
        return header + value.packed



from ._exceptions import ProgrammingError


class Cursor:
    def __init__(self, stmt):
        self.stmt = stmt

    async def _init(self):
        if not self.stmt.conn.in_transaction:
            raise ProgrammingError(
                'Cursors can only be used inside transactions')

    async def fetch_row(self):
        rows = await self.fetch(1)
        if rows:
            return rows[0]
        else:
            return None

    async def fetch(self, n):
        if self.stmt.finished:
            return []
        return await self.stmt.exec_continue(limit=n)

    async def forward(self, n):
        if self.stmt.finished:
            return
        return await self.stmt.forward(n)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.stmt.finished:
            raise StopAsyncIteration

        result = await self.fetch_row()
        if result:
            return result

        raise StopAsyncIteration

    def __iter__(self):
        raise TypeError(
            'Cursor object can only be iterated asynchronously. '
            'Did you forget to use "async for"?')



from setuptools import setup

with open('pgtrio/_version.py') as f:
    exec(f.read())

test_dependencies = [
    'pytest==7.1.1',
    'pytest-trio==0.7.0',
]

dev_dependencies = test_dependencies + [
    'wheel',
    'twine==4.0.0',
]

setup(
    name='pgtrio',
    version=__version__,
    description='A Trio-Native PostgreSQL Interface Library',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='http://github.com/elektito/pgtrio',
    author='Mostafa Razavi',
    author_email='mostafa@sepent.com',
    license='MIT',
    packages=['pgtrio'],
    zip_safe=False,
    install_requires=[
        'trio==0.20.0',
        'python-dateutil==2.8.2',
        'orjson==3.6.8',
    ],
    extras_require={
        'test': test_dependencies,
        'dev': dev_dependencies,
    }
)
